from __future__ import annotations

import asyncio
import hashlib
import inspect
import json
import logging
import math
import os
import re
import string
from collections.abc import Iterable
from datetime import datetime
from functools import reduce
from http import HTTPStatus
from pathlib import Path
from typing import Any, BinaryIO, Collection, Coroutine, Iterator, Union
from uuid import UUID

import aiohttp
import httpx
import pypdf
from pybtex.database import Person, parse_string
from pybtex.database.input.bibtex import Parser
from pybtex.style.formatting import unsrtalpha
from pybtex.style.template import FieldIsMissing
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception,
    stop_after_attempt,
    wait_incrementing,
)

logger = logging.getLogger(__name__)


StrPath = Union[str, Path]


def name_in_text(name: str, text: str) -> bool:
    sname = name.strip()
    pattern = rf"\b({re.escape(sname)})\b(?!\w)"
    return bool(re.search(pattern, text))


def maybe_is_text(s: str, thresh: float = 2.5) -> bool:
    if len(s) == 0:
        return False
    # Calculate the entropy of the string
    entropy = 0.0
    for c in string.printable:
        p = s.count(c) / len(s)
        if p > 0:
            entropy += -p * math.log2(p)

    # Check if the entropy is within a reasonable range for text
    return entropy > thresh


def maybe_is_pdf(file: BinaryIO) -> bool:
    magic_number = file.read(4)
    file.seek(0)
    return magic_number == b"%PDF"


def maybe_is_html(file: BinaryIO) -> bool:
    magic_number = file.read(4)
    file.seek(0)
    return magic_number in {b"<htm", b"<!DO", b"<xsl", b"<!X"}


def strings_similarity(s1: str, s2: str) -> float:
    if len(s1) == 0 or len(s2) == 0:
        return 0
    # break the strings into words
    ss1 = set(s1.split())
    ss2 = set(s2.split())
    # return the similarity ratio
    return len(ss1.intersection(ss2)) / len(ss1.union(ss2))


def count_pdf_pages(file_path: StrPath) -> int:
    with open(file_path, "rb") as pdf_file:
        try:  # try fitz by default
            import fitz

            doc = fitz.open(file_path)
            num_pages = len(doc)
        except ModuleNotFoundError:  # pypdf instead
            pdf_reader = pypdf.PdfReader(pdf_file)
            num_pages = len(pdf_reader.pages)
    return num_pages


def hexdigest(data: str | bytes) -> str:
    if isinstance(data, str):
        return hashlib.md5(data.encode("utf-8")).hexdigest()  # noqa: S324
    return hashlib.md5(data).hexdigest()  # noqa: S324


def md5sum(file_path: StrPath) -> str:
    with open(file_path, "rb") as f:
        return hexdigest(f.read())


async def gather_with_concurrency(n: int, coros: list[Coroutine]) -> list[Any]:
    # https://stackoverflow.com/a/61478547/2392535
    semaphore = asyncio.Semaphore(n)

    async def sem_coro(coro):
        async with semaphore:
            return await coro

    return await asyncio.gather(*(sem_coro(c) for c in coros))


def strip_citations(text: str) -> str:
    # Combined regex for identifying citations (see unit tests for examples)
    citation_regex = r"\b[\w\-]+\set\sal\.\s\([0-9]{4}\)|\((?:[^\)]*?[a-zA-Z][^\)]*?[0-9]{4}[^\)]*?)\)"
    # Remove the citations from the text
    return re.sub(citation_regex, "", text, flags=re.MULTILINE)


def get_citenames(text: str) -> set[str]:
    # Combined regex for identifying citations (see unit tests for examples)
    citation_regex = r"\b[\w\-]+\set\sal\.\s\([0-9]{4}\)|\((?:[^\)]*?[a-zA-Z][^\)]*?[0-9]{4}[^\)]*?)\)"
    results = re.findall(citation_regex, text, flags=re.MULTILINE)
    # now find None patterns
    none_citation_regex = r"(\(None[a-f]{0,1} pages [0-9]{1,10}-[0-9]{1,10}\))"
    none_results = re.findall(none_citation_regex, text, flags=re.MULTILINE)
    results.extend(none_results)
    values = []
    for citation in results:
        citation = citation.strip("() ")  # noqa: PLW2901
        for c in re.split(",|;", citation):
            if c == "Extra background information":
                continue
            # remove leading/trailing spaces
            c = c.strip()  # noqa: PLW2901
            values.append(c)
    return set(values)


def extract_doi(reference: str) -> str:
    """
    Extracts DOI from the reference string using regex.

    :param reference: A string containing the reference.
    :return: A string containing the DOI link or a message if DOI is not found.
    """
    # DOI regex pattern
    doi_pattern = r"10.\d{4,9}/[-._;()/:A-Z0-9]+"
    doi_match = re.search(doi_pattern, reference, re.IGNORECASE)

    # If DOI is found in the reference, return the DOI link
    if doi_match:
        return "https://doi.org/" + doi_match.group()
    else:  # noqa: RET505
        return ""


def batch_iter(iterable: list, n: int = 1) -> Iterator[list]:
    """
    Batch an iterable into chunks of size n.

    :param iterable: The iterable to batch
    :param n: The size of the batches
    :return: A list of batches
    """
    length = len(iterable)
    for ndx in range(0, length, n):
        yield iterable[ndx : min(ndx + n, length)]


def flatten(iteratble: list) -> list:
    """
    Flatten a list of lists.

    :param l: The list of lists to flatten
    :return: A flattened list
    """
    return [item for sublist in iteratble for item in sublist]


def get_loop() -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def is_coroutine_callable(obj):
    if inspect.isfunction(obj):
        return inspect.iscoroutinefunction(obj)
    elif callable(obj):  # noqa: RET505
        return inspect.iscoroutinefunction(obj.__call__)
    return False


def llm_read_json(text: str) -> dict:
    """Read LLM output and extract JSON data from it."""
    # fetch from markdown ```json if present
    text = text.strip().split("```json")[-1].split("```")[0]
    # split anything before the first {
    text = "{" + text.split("{", 1)[-1]
    # split anything after the last }
    text = text.rsplit("}", 1)[0] + "}"

    # escape new lines within strings
    def replace_newlines(match: re.Match) -> str:
        return match.group(0).replace("\n", "\\n")

    # Match anything between double quotes
    # including escaped quotes and other escaped characters.
    # https://regex101.com/r/VFcDmB/1
    pattern = r'"(?:[^"\\]|\\.)*"'
    text = re.sub(pattern, replace_newlines, text)

    return json.loads(text)


def encode_id(value: str | bytes | UUID, maxsize: int | None = 16) -> str:
    """Encode a value (e.g. a DOI) optionally with a max length."""
    if isinstance(value, UUID):
        value = str(value)
    if isinstance(value, str):
        value = value.lower().encode()
    return hashlib.md5(value).hexdigest()[:maxsize]  # noqa: S324


def get_year(ts: datetime | None = None) -> str:
    """Get the year from the input datetime, otherwise using the current datetime."""
    if ts is None:
        ts = datetime.now()
    return ts.strftime("%Y")


class CitationConversionError(Exception):
    """Exception to throw when we can't process a citation from a BibTeX."""


def clean_upbibtex(bibtex: str) -> str:

    if not bibtex:
        return bibtex

    mapping = {
        "None": "article",
        "Article": "article",
        "JournalArticle": "article",
        "Review": "article",
        "Book": "book",
        "BookSection": "inbook",
        "ConferencePaper": "inproceedings",
        "Conference": "inproceedings",
        "Dataset": "misc",
        "Dissertation": "phdthesis",
        "Journal": "article",
        "Patent": "patent",
        "Preprint": "article",
        "Report": "techreport",
        "Thesis": "phdthesis",
        "WebPage": "misc",
        "Plain": "article",
    }
    if "@None" in bibtex:
        return bibtex.replace("@None", "@article")
    match = re.findall(r"@\['(.*)'\]", bibtex)
    if not match:
        match = re.findall(r"@(\w+)\{", bibtex)
        bib_type = match[0]
        current = f"@{match[0]}"
    else:
        bib_type = match[0]
        current = f"@['{bib_type}']"
    for k, v in mapping.items():
        # can have multiple
        if k in bib_type:
            bibtex = bibtex.replace(current, f"@{v}")
            break
    return bibtex


def format_bibtex(  # noqa: C901
    bibtex: str,
    key: str | None = None,
    clean: bool = True,
    missing_replacements: dict[str, str] | None = None,
) -> str:
    """Transform bibtex entry into a citation, potentially adding missing fields."""
    if missing_replacements is None:
        missing_replacements = {}
    if key is None:
        key = bibtex.split("{")[1].split(",")[0]
    style = unsrtalpha.Style()
    try:
        bd = parse_string(clean_upbibtex(bibtex) if clean else bibtex, "bibtex")
    except Exception:
        return "Ref " + key
    try:
        entry = bd.entries[key]
    except KeyError as exc:  # Let's check if key is a non-empty prefix
        try:
            entry = next(
                iter(v for k, v in bd.entries.items() if k.startswith(key) and key)
            )
        except StopIteration:
            raise CitationConversionError(
                f"Failed to process{' and clean up' if clean else ''} bibtex {bibtex}"
                f" due to failed lookup of key {key}."
            ) from exc
    try:
        # see if we can insert missing fields
        for field, replacement_value in missing_replacements.items():
            # Deal with special case for author, since it needs to be parsed
            # into Person objects. This reorganizes the names automatically.
            if field == "author" and "author" not in entry.persons:
                tmp_author_bibtex = f"@misc{{tmpkey, author={{{replacement_value}}}}}"
                authors: list[Person] = (
                    Parser()
                    .parse_string(tmp_author_bibtex)
                    .entries["tmpkey"]
                    .persons["author"]
                )
                for a in authors:
                    entry.add_person(a, "author")
            elif field not in entry.fields:
                entry.fields.update({field: replacement_value})
        entry = style.format_entry(label="1", entry=entry)
        return entry.text.render_as("text")
    except (FieldIsMissing, UnicodeDecodeError):
        try:
            return entry.fields["title"]
        except KeyError as exc:
            raise CitationConversionError(
                f"Failed to process{' and clean up' if clean else ''} bibtex {bibtex}"
                " due to missing a 'title' field."
            ) from exc


def remove_substrings(target: str, substr_removal_list: Collection[str]) -> str:
    """Remove substrings from a target string."""
    if all(len(w) == 1 for w in substr_removal_list):
        return target.translate(str.maketrans("", "", "".join(substr_removal_list)))

    for substr in substr_removal_list:
        target = target.replace(substr, "")
    return target


def bibtex_field_extract(
    bibtex: str, field: str, missing_replacements: dict[str, str] | None = None
) -> str:
    """Get a field from a bibtex entry.

    Args:
        bibtex: bibtex entry
        field: field to extract
        missing_replacements: replacement extract for field if not present in the bibtex string
    """
    if missing_replacements is None:
        missing_replacements = {}
    try:
        pattern = rf"{field}\s*=\s*{{(.*?)}},"
        # note: we intentionally have an attribute error if no match
        return re.search(pattern, bibtex, re.IGNORECASE).group(1).strip()  # type: ignore[union-attr]
    except AttributeError:
        return missing_replacements.get(field, "")


def create_bibtex_key(author: list[str], year: str, title: str) -> str:
    FORBIDDEN_KEY_CHARACTERS = {"_", " ", "-", "/", "'", "`", ":", ",", "\n"}
    author_rep = (
        author[0].split()[-1].casefold()
        if "Unknown" not in author[0]
        else "unknownauthors"
    )
    # we don't want a bibtex-parsing induced line break in the key
    # so we cap it to 100+50+4 = 154 characters max
    # 50 for the author, 100 for the first three title words, 4 for the year
    # the first three title words are just emulating the s2 convention
    key = f"{author_rep[:50]}{year}{''.join([t.casefold() for t in title.split()[:3]])[:100]}"
    return remove_substrings(key, FORBIDDEN_KEY_CHARACTERS)


@retry(
    retry=retry_if_exception(
        lambda x: isinstance(x, aiohttp.ServerDisconnectedError)
        or isinstance(x, aiohttp.ClientResponseError)
        and x.status
        in {
            httpx.codes.INTERNAL_SERVER_ERROR.value,
            httpx.codes.GATEWAY_TIMEOUT.value,
        }
    ),
    before_sleep=before_sleep_log(logger, logging.WARNING),
    stop=stop_after_attempt(5),
    wait=wait_incrementing(0.1, 0.1),
)
async def _get_with_retrying(
    url: str,
    params: dict[str, Any],
    session: aiohttp.ClientSession,
    headers: dict[str, str] | None = None,
    timeout: float = 10.0,  # noqa: ASYNC109
    http_exception_mappings: dict[HTTPStatus | int, Exception] | None = None,
) -> dict[str, Any]:
    """Get from a URL with retrying protection."""
    try:
        async with session.get(
            url,
            params=params,
            headers=headers,
            timeout=aiohttp.ClientTimeout(timeout),
        ) as response:
            response.raise_for_status()
            return await response.json()
    except aiohttp.ClientResponseError as e:
        if http_exception_mappings and e.status in http_exception_mappings:
            raise http_exception_mappings[e.status] from e
        raise


def union_collections_to_ordered_list(collections: Iterable) -> list:
    return sorted(reduce(lambda x, y: set(x) | set(y), collections))


def pqa_directory(name: str) -> Path:
    if pqa_home := os.environ.get("PQA_HOME"):
        directory = Path(pqa_home) / ".pqa" / name
    else:
        directory = Path.home() / ".pqa" / name

    directory.mkdir(parents=True, exist_ok=True)
    return directory



from __future__ import annotations

import itertools
import json
import logging
import re
from pathlib import Path
from typing import Any, cast
from unittest.mock import patch

import anyio
import pytest
from pydantic import ValidationError
from pytest_subtests import SubTests

from paperqa.docs import Docs
from paperqa.llms import LangchainLLMModel
from paperqa.types import Answer, Context, Doc, PromptCollection, Text
from paperqa.utils import get_year

try:
    from langchain.agents.openai_functions_agent.base import OpenAIFunctionsAgent
    from langchain_openai import ChatOpenAI
    from tenacity import Retrying, retry_if_exception_type, stop_after_attempt

    from paperqa.agents import agent_query
    from paperqa.agents.helpers import (
        compute_total_model_token_cost,
        update_doc_models,
    )
    from paperqa.agents.models import (
        AgentPromptCollection,
        AgentStatus,
        AnswerResponse,
        MismatchedModelsError,
        QueryRequest,
    )
    from paperqa.agents.prompts import STATIC_PROMPTS
    from paperqa.agents.search import get_directory_index
    from paperqa.agents.tools import (
        GatherEvidenceTool,
        GenerateAnswerTool,
        PaperSearchTool,
        SharedToolState,
    )
except ImportError:
    pytest.skip("agents module is not installed", allow_module_level=True)


PAPER_DIRECTORY = Path(__file__).parent


@pytest.mark.asyncio
async def test_get_directory_index(agent_index_dir):
    index = await get_directory_index(
        directory=anyio.Path(PAPER_DIRECTORY),
        index_name="pqa_index_0",
        index_directory=agent_index_dir,
    )
    assert index.fields == [
        "title",
        "file_location",
        "body",
        "year",
    ], "Incorrect fields in index"
    assert len(await index.index_files) == 4, "Incorrect number of index files"
    results = await index.query(query="who is Frederick Bates?")
    # added docs.keys come from md5 hash of the file location
    assert results[0].docs.keys() == {"dab5b86dea3bd4c7ffe05a9f33ae95f7"}


@pytest.mark.asyncio
async def test_get_directory_index_w_manifest(agent_index_dir):
    index = await get_directory_index(
        directory=anyio.Path(PAPER_DIRECTORY),
        index_name="pqa_index_0",
        index_directory=agent_index_dir,
        manifest_file=anyio.Path(PAPER_DIRECTORY) / "stub_manifest.csv",
    )
    assert index.fields == [
        "title",
        "file_location",
        "body",
        "year",
    ], "Incorrect fields in index"
    # 4 = example.txt + example2.txt + paper.pdf + example.html
    assert len(await index.index_files) == 4, "Incorrect number of index files"
    results = await index.query(query="who is Barack Obama?")
    top_result = next(iter(results[0].docs.values()))
    assert top_result.dockey == "af2c9acf6018e62398fc6efc4f0a04b4"
    # note: this title comes from the manifest, so we know it worked
    assert top_result.title == "Barack Obama (Wikipedia article)"


@pytest.mark.flaky(reruns=3, only_rerun=["AssertionError", "httpx.RemoteProtocolError"])
@pytest.mark.parametrize("agent_type", ["OpenAIFunctionsAgent", "fake"])
@pytest.mark.asyncio
async def test_agent_types(agent_index_dir, agent_type):

    question = "How can you use XAI for chemical property prediction?"

    request = QueryRequest(
        query=question,
        consider_sources=10,
        max_sources=2,
        embedding="sparse",
        agent_tools=AgentPromptCollection(
            search_count=2,
            paper_directory=PAPER_DIRECTORY,
            index_directory=agent_index_dir,
        ),
    )
    response = await agent_query(
        request, agent_type=agent_type, index_directory=agent_index_dir
    )
    assert response.answer.answer != "I cannot answer", "Answer not generated"
    assert len(response.answer.context) >= 1, "No contexts were found"
    assert response.answer.question == question


@pytest.mark.asyncio
async def test_timeout(agent_index_dir):
    response = await agent_query(
        QueryRequest(
            query="Are COVID-19 vaccines effective?",
            llm="gpt-4o-mini",
            prompts=PromptCollection(pre=None),
            # We just need one tool to test the timeout, gen_answer is not that fast
            agent_tools=AgentPromptCollection(
                timeout=0.001,
                tool_names={"gen_answer"},
                paper_directory=PAPER_DIRECTORY,
                index_directory=agent_index_dir,
            ),
        ),
        Docs(),
    )
    # ensure that GenerateAnswerTool was called
    assert response.status == AgentStatus.TIMEOUT, "Agent did not timeout"
    assert "I cannot answer" in response.answer.answer


@pytest.mark.asyncio
async def test_propagate_options(agent_index_dir) -> None:
    llm_name = "gpt-4o-mini"
    default_llm_names = {
        cls.model_fields[name].default  # type: ignore[attr-defined]
        for name, cls in itertools.product(("llm", "summary_llm"), (QueryRequest, Docs))
    }
    assert (
        llm_name not in default_llm_names
    ), f"Assertions require not matching a default LLM name in {default_llm_names}."
    query = QueryRequest(
        query="What is is a self-explanatory model?",
        summary_llm=llm_name,
        llm=llm_name,
        max_sources=5,
        consider_sources=6,
        length="400 words",
        prompts=PromptCollection(
            pre=None, system="End all responses with ###", skip_summary=True
        ),
        # NOTE: this is testing that if our prompt forgets template fields (e.g. status),
        # the code still runs, despite the presence of extra keyword arguments to format
        agent_tools=AgentPromptCollection(
            paper_directory=PAPER_DIRECTORY,
            index_directory=agent_index_dir,
            agent_prompt=(
                "Answer question: {question}. Search for papers, gather evidence, and"
                " answer. If you do not have enough evidence, you can search for more"
                " papers (preferred) or gather more evidence with a different phrase."
                " You may rephrase or break-up the question in those steps. Once you"
                " have five or more pieces of evidence from multiple sources, or you"
                " have tried a few times, call {gen_answer_tool_name} tool. The"
                " {gen_answer_tool_name} tool output is visible to the user, so you do"
                " not need to restate the answer and can simply terminate if the answer"
                " looks sufficient."
            ),
            tool_names={"paper_search", "gather_evidence", "gen_answer"},
        ),
    )
    for attempt in Retrying(
        stop=stop_after_attempt(3), retry=retry_if_exception_type(AssertionError)
    ):
        with attempt:
            docs = Docs(llm=llm_name, summary_llm=llm_name)
            docs.prompts = query.prompts  # this line happens in main
            response = await agent_query(query, docs, agent_type="fake")
            assert response.status == AgentStatus.SUCCESS, "Agent did not succeed"
    result = response.answer
    assert len(result.answer) > 200, "Answer did not return any results"
    assert result.answer_length == query.length, "Answer length did not propagate"
    assert "###" in result.answer, "Answer did not propagate system prompt"


@pytest.mark.flaky(reruns=3, only_rerun=["AssertionError"])
@pytest.mark.asyncio
async def test_mixing_langchain_clients(caplog, agent_index_dir) -> None:
    docs = Docs()
    query = QueryRequest(
        query="What is is a self-explanatory model?",
        max_sources=2,
        consider_sources=3,
        llm="gemini-1.5-flash",
        summary_llm="gemini-1.5-flash",
        agent_tools=AgentPromptCollection(
            paper_directory=PAPER_DIRECTORY, index_directory=agent_index_dir
        ),
    )
    update_doc_models(docs, query)
    with caplog.at_level(logging.WARNING):
        response = await agent_query(query, docs)
    assert response.status == AgentStatus.SUCCESS, "Agent did not succeed"
    assert not [
        msg for (*_, msg) in caplog.record_tuples if "error" in msg.lower()
    ], "Expected clean logs"


@pytest.mark.asyncio
async def test_gather_evidence_rejects_empty_docs() -> None:
    # Patch GenerateAnswerTool._arun so that if this tool is chosen first, we
    # don't give a 'cannot answer' response. A 'cannot answer' response can
    # lead to an unsure status, which will break this test's assertions. Since
    # this test is about a GatherEvidenceTool edge case, defeating
    # GenerateAnswerTool is fine
    with patch.object(
        GenerateAnswerTool, "_arun", return_value="Failed to answer question."
    ):
        response = await agent_query(
            query=QueryRequest(
                query="Are COVID-19 vaccines effective?",
                agent_tools=AgentPromptCollection(
                    tool_names={"gather_evidence", "gen_answer"}
                ),
            ),
            docs=Docs(),
        )
    assert response.status == AgentStatus.FAIL, "Agent should have registered a failure"


@pytest.mark.flaky(reruns=3, only_rerun=["AssertionError"])
@pytest.mark.asyncio
async def test_agent_sharing_state(
    fixture_stub_answer, subtests: SubTests, agent_index_dir
) -> None:
    tool_state = SharedToolState(docs=Docs(), answer=fixture_stub_answer)
    search_count = 3  # Keep low for speed
    query = QueryRequest(
        query=fixture_stub_answer.question,
        consider_sources=2,
        max_sources=1,
        agent_tools=AgentPromptCollection(
            search_count=search_count,
            index_directory=agent_index_dir,
            paper_directory=PAPER_DIRECTORY,
        ),
    )

    with subtests.test(msg=PaperSearchTool.__name__):
        tool = PaperSearchTool(
            shared_state=tool_state,
            search_count=search_count,
            index_directory=agent_index_dir,
            paper_directory=PAPER_DIRECTORY,
        )
        await tool.arun("XAI self explanatory model")
        assert tool_state.docs.docs, "Search did not save any papers"
        assert all(
            (isinstance(d, Doc) or issubclass(d, Doc))  # type: ignore[unreachable]
            for d in tool_state.docs.docs.values()
        ), "Document type or DOI propagation failure"

    with subtests.test(msg=GatherEvidenceTool.__name__):
        assert (
            not fixture_stub_answer.contexts
        ), "No contexts is required for a later assertion"

        tool = GatherEvidenceTool(shared_state=tool_state, query=query)
        await tool.arun(fixture_stub_answer.question)
        assert (
            len(fixture_stub_answer.dockey_filter) > 0
        ), "Filter did not preserve reference"
        assert fixture_stub_answer.contexts, "Evidence did not return any results"

    with subtests.test(msg=f"{GenerateAnswerTool.__name__} working"):
        tool = GenerateAnswerTool(shared_state=tool_state, query=query)
        result = await tool.arun(fixture_stub_answer.question)
        assert re.search(
            pattern=SharedToolState.STATUS_SEARCH_REGEX_PATTERN, string=result
        )
        assert (
            len(fixture_stub_answer.answer) > 200
        ), "Answer did not return any results"
        assert (
            GenerateAnswerTool.extract_answer_from_message(result)
            == fixture_stub_answer.answer
        ), "Failed to regex extract answer from result"
        assert (
            len(fixture_stub_answer.contexts) <= query.max_sources
        ), "Answer has more sources than expected"

    with subtests.test(msg=f"{GenerateAnswerTool.__name__} misconfigured query"):
        query.consider_sources = 1  # k
        query.max_sources = 5
        tool = GenerateAnswerTool(shared_state=tool_state, query=query)
        with pytest.raises(ValueError, match="k should be greater than max_sources"):
            await tool.arun("Are COVID-19 vaccines effective?")


def test_functions() -> None:
    """Check the functions schema passed to OpenAI."""
    shared_tool_state = SharedToolState(
        answer=Answer(question="stub"),
        docs=Docs(),
    )
    agent = OpenAIFunctionsAgent.from_llm_and_tools(
        llm=ChatOpenAI(model="gpt-4-turbo-2024-04-09"),
        tools=[
            PaperSearchTool(shared_state=shared_tool_state),
            GatherEvidenceTool(shared_state=shared_tool_state, query=QueryRequest()),
            GenerateAnswerTool(shared_state=shared_tool_state, query=QueryRequest()),
        ],
    )
    assert cast(OpenAIFunctionsAgent, agent).functions == [
        {
            "name": "paper_search",
            "description": (
                "Search for papers to increase the paper count. You can call this a second "
                "time with an different search to gather more papers."
            ),
            "parameters": {
                "type": "object",
                "properties": {
                    "query": {
                        "description": (
                            "A search query in this format: [query], [start year]-[end year]. "
                            "You may include years as the last word in the query, "
                            "e.g. 'machine learning 2020' or 'machine learning 2010-2020'. "
                            f"The current year is {get_year()}. "
                            "The query portion can be a specific phrase, complete sentence, "
                            "or general keywords, e.g. 'machine learning for immunology'."
                        ),
                        "type": "string",
                    }
                },
                "required": ["query"],
            },
        },
        {
            "name": "gather_evidence",
            "description": (
                "Gather evidence from previous papers given a specific question. "
                "This will increase evidence and relevant paper counts. "
                "Only invoke when paper count is above zero."
            ),
            "parameters": {
                "type": "object",
                "properties": {
                    "question": {
                        "description": "Specific question to gather evidence for.",
                        "type": "string",
                    }
                },
                "required": ["question"],
            },
        },
        {
            "name": "gen_answer",
            "description": (
                "Ask a model to propose an answer answer using current evidence. "
                "The tool may fail, "
                "indicating that better or different evidence should be found. "
                "Having more than one piece of evidence or relevant papers is best."
            ),
            "parameters": {
                "type": "object",
                "properties": {
                    "question": {
                        "description": "Question to be answered.",
                        "type": "string",
                    }
                },
                "required": ["question"],
            },
        },
    ]


def test_instruct_model():
    docs = Docs(name="tmp")
    query = QueryRequest(
        summary_llm="gpt-3.5-turbo-instruct",
        query="Are COVID-19 vaccines effective?",
        llm="gpt-3.5-turbo-instruct",
    )
    update_doc_models(docs, query)
    docs.query("Are COVID-19 vaccines effective?")
    query.llm = "gpt-3.5-turbo-instruct"
    query.summary_llm = "gpt-3.5-turbo-instruct"
    update_doc_models(docs, query)
    docs.query("Are COVID-19 vaccines effective?")


def test_anthropic_model():
    docs = Docs(name="tmp")
    query = QueryRequest(
        summary_llm="claude-3-sonnet-20240229",
        query="Are COVID-19 vaccines effective?",
        llm="claude-3-sonnet-20240229",
    )
    update_doc_models(docs, query)
    answer = docs.query("Are COVID-19 vaccines effective?")

    # make sure we can compute cost with this model.
    compute_total_model_token_cost(answer.token_counts)


def test_embeddings_anthropic():
    docs = Docs(name="tmp")
    query = QueryRequest(
        summary_llm="claude-3-sonnet-20240229",
        query="Are COVID-19 vaccines effective?",
        llm="claude-3-sonnet-20240229",
        embedding="sparse",
    )
    update_doc_models(docs, query)
    _ = docs.query("Are COVID-19 vaccines effective?")

    query = QueryRequest(
        summary_llm="claude-3-sonnet-20240229",
        query="Are COVID-19 vaccines effective?",
        llm="claude-3-sonnet-20240229",
        embedding="hybrid-text-embedding-3-small",
    )
    update_doc_models(docs, query)
    _ = docs.query("Are COVID-19 vaccines effective?")


@pytest.mark.asyncio
async def test_gemini_model_construction(
    stub_paper_path: Path,
) -> None:
    docs = Docs(name="tmp")
    query = QueryRequest(
        summary_llm="gemini-1.5-pro",
        llm="gemini-1.5-pro",
        embedding="sparse",
    )
    update_doc_models(docs, query)
    assert isinstance(docs.llm_model, LangchainLLMModel)  # We use LangChain for Gemini
    assert docs.llm_model.name == "gemini-1.5-pro", "Gemini Model: model not created"
    assert "model" not in docs.llm_model.config, "model should not be in config"

    # now try using it
    await docs.aadd(stub_paper_path)
    answer = await docs.aget_evidence(
        Answer(question="Are COVID-19 vaccines effective?")
    )
    assert answer.contexts, "Gemini Model: no contexts returned"


def test_query_request_summary():
    """Test that we can set summary llm to none and it will skip summary."""
    request = QueryRequest(query="Are COVID-19 vaccines effective?", summary_llm="none")
    assert request.summary_llm == "gpt-4o-mini"
    assert request.prompts.skip_summary, "Summary should be skipped with none llm"


def test_query_request_preset_prompts():
    """Test that we can set the prompt using our preset prompts."""
    request = QueryRequest(
        query="Are COVID-19 vaccines effective?",
        prompts=PromptCollection(
            summary_json_system=r"{gene_name} {summary} {relevance_score}"
        ),
    )
    assert "gene_name" in request.prompts.summary_json_system


def test_query_request_docs_name_serialized():
    """Test that the query request has a docs_name property."""
    request = QueryRequest(query="Are COVID-19 vaccines effective?")
    request_data = json.loads(request.model_dump_json())
    assert "docs_name" in request_data
    assert request_data["docs_name"] is None
    request.set_docs_name("my_doc")
    request_data = json.loads(request.model_dump_json())
    assert request_data["docs_name"] == "my_doc"


def test_query_request_model_mismatch():
    with pytest.raises(
        MismatchedModelsError,
        match=(
            "Answer LLM and summary LLM types must match: "
            "<class 'paperqa.llms.AnthropicLLMModel'> != <class 'paperqa.llms.OpenAILLMModel'>"
        ),
    ):
        _ = QueryRequest(
            summary_llm="gpt-4o-mini",
            query="Are COVID-19 vaccines effective?",
            llm="claude-3-sonnet-20240229",
            embedding="sparse",
        )


def test_answers_are_striped():
    """Test that answers are striped."""
    answer = Answer(
        question="What is the meaning of life?",
        contexts=[
            Context(
                context="bla",
                text=Text(
                    name="text",
                    text="The meaning of life is 42.",
                    embedding=[43.3, 34.2],
                    doc=Doc(
                        docname="foo",
                        citation="bar",
                        dockey="baz",
                        embedding=[43.1, 65.2],
                    ),
                ),
                score=3,
            )
        ],
    )
    response = AnswerResponse(answer=answer, usage={}, bibtex={}, status="success")

    assert response.answer.contexts[0].text.embedding is None
    assert response.answer.contexts[0].text.text == ""  # type: ignore[unreachable,unused-ignore]
    assert response.answer.contexts[0].text.doc is not None
    assert response.answer.contexts[0].text.doc.embedding is None
    # make sure it serializes
    response.model_dump_json()


def test_prompts_are_set():
    assert (
        STATIC_PROMPTS["json"].summary_json_system
        != PromptCollection().summary_json_system
    )


@pytest.mark.parametrize(
    ("kwargs", "result"),
    [
        ({}, None),
        ({"tool_names": {GenerateAnswerTool.__fields__["name"].default}}, None),
        ({"tool_names": set()}, ValidationError),
        ({"tool_names": {PaperSearchTool.__fields__["name"].default}}, ValidationError),
    ],
)
def test_agent_prompt_collection_validations(
    kwargs: dict[str, Any], result: type[Exception] | None
) -> None:
    if result is None:
        AgentPromptCollection(**kwargs)
    else:
        with pytest.raises(result):
            AgentPromptCollection(**kwargs)



from __future__ import annotations
import inspect
import logging
import os
import re
import sys
from typing import ClassVar
import anyio
from langchain_core.callbacks import BaseCallbackHandler

from langchain.tools import BaseTool
from paperqa import Answer, Docs
from ..utils import pqa_directory
from pydantic import BaseModel, ConfigDict, Field

# ruff: noqa: I001
from pydantic.v1 import (  # TODO: move to Pydantic v2 after LangChain moves to Pydantic v2, SEE: https://github.com/langchain-ai/langchain/issues/16306
    BaseModel as BaseModelV1,
    Field as FieldV1,
)

from .helpers import compute_total_model_token_cost, get_year
from .search import get_directory_index
from .models import ParsingConfiguration, QueryRequest, SimpleProfiler

logger = logging.getLogger(__name__)


async def status(docs: Docs, answer: Answer, relevant_score_cutoff: int = 5) -> str:
    """Create a string that provides a summary of the input doc/answer."""
    answer.cost = compute_total_model_token_cost(answer.token_counts)
    return (
        f"Status: Paper Count={len(docs.docs)}"
        f" | Relevant Papers={len({c.text.doc.dockey for c in answer.contexts if c.score > relevant_score_cutoff})}"
        f" | Current Evidence={len([c for c in answer.contexts if c.score > relevant_score_cutoff])}"
        f" | Current Cost=${answer.cost:.2f}"
    )


class SharedToolState(BaseModel):
    """Shared collection of variables for collection of tools. We use this to avoid
    the fact that pydantic treats dictionaries as values, instead of references.
    """  # noqa: D205

    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    answer: Answer
    docs: Docs
    profiler: SimpleProfiler = Field(default_factory=SimpleProfiler)

    # SEE: https://regex101.com/r/RmuVdC/1
    STATUS_SEARCH_REGEX_PATTERN: ClassVar[str] = (
        r"Status: Paper Count=(\d+) \| Relevant Papers=(\d+) \| Current Evidence=(\d+)"
    )

    async def get_status(self) -> str:
        return await status(self.docs, self.answer)


def _time_tool(func):
    """Decorator to time the execution of a tool method.
    Assumes that the tool has a shared state.
    """  # noqa: D205

    async def wrapper(self, *args, **kwargs):
        async with self.shared_state.profiler.timer(self.name):
            return await func(self, *args, **kwargs)

    return wrapper


class PaperSearchTool(BaseTool):
    class InputSchema(
        BaseModelV1  # TODO: move to Pydantic v2 after LangChain moves to Pydantic v2, SEE: https://github.com/langchain-ai/langchain/issues/16306
    ):
        query: str = FieldV1(
            description=(
                "A search query in this format: [query], [start year]-[end year]. "
                "You may include years as the last word in the query, "
                "e.g. 'machine learning 2020' or 'machine learning 2010-2020'. "
                f"The current year is {get_year()}. "
                "The query portion can be a specific phrase, complete sentence, "
                "or general keywords, e.g. 'machine learning for immunology'."
            )
        )

    paper_directory: str | os.PathLike = "."
    index_directory: str | os.PathLike | None = None
    manifest_file: str | os.PathLike | None = None
    name: str = "paper_search"
    args_schema: type[BaseModelV1] | None = InputSchema
    description: str = (
        "Search for papers to increase the paper count. You can call this a second "
        "time with an different search to gather more papers."
    )

    shared_state: SharedToolState
    return_paper_metadata: bool = False
    # Second item being True means specify a year range in the search
    search_type: tuple[str, bool] = ("google", False)
    search_count: int = 8
    previous_searches: dict[str, int] = FieldV1(default_factory=dict)
    embedding: str = "text-embedding-3-small"
    parsing_configuration: ParsingConfiguration = FieldV1(
        default_factory=ParsingConfiguration
    )

    def _run(self, query: str) -> str:
        raise NotImplementedError

    @_time_tool
    async def _arun(self, query: str) -> str:
        """
        Run asynchronously, in-place mutating `self.shared_state.docs`.

        Args:
            query: Search keywords followed by optional year or year range
                (e.g. COVID-19 vaccines, 2022).

        Returns:
            String describing searched papers and the current status.
        """
        # get offset if we've done this search before (continuation of search)
        # or mark this search as new (so offset 0)
        logger.info(f"Starting paper search for '{query}'.")
        search_key = query
        if search_key in self.previous_searches:
            offset = self.previous_searches[search_key]
        else:
            offset = self.previous_searches[search_key] = 0

        # Preprocess inputs to make ScrapeRequest
        keywords = query.replace('"', "")  # Remove quotes
        year: str | None = None
        last_word = keywords.split(" ")[-1]
        if re.match(r"\d{4}(-\d{4})?", last_word):
            keywords = keywords[: -len(last_word)].removesuffix(",").strip()
            if self.search_type[1]:
                year = last_word
                if "-" not in year:
                    year = year + "-" + year  # Convert to date range (e.g. 2022-2022)
        index = await get_directory_index(
            directory=anyio.Path(self.paper_directory),
            index_name=QueryRequest.get_index_name(
                self.paper_directory, self.embedding, self.parsing_configuration
            ),
            index_directory=self.index_directory,
            manifest_file=(
                anyio.Path(self.manifest_file) if self.manifest_file else None
            ),
            embedding=self.embedding,
            chunk_chars=self.parsing_configuration.chunksize,
            overlap=self.parsing_configuration.overlap,
        )

        results = await index.query(
            keywords,
            top_n=self.search_count,
            offset=offset,
            field_subset=[f for f in index.fields if f != "year"],
        )

        logger.info(f"Search for '{keywords}' returned {len(results)} papers.")
        # combine all the resulting doc objects into one and update the state
        # there's only one doc per result, so we can just take the first one
        all_docs = []
        for r in results:
            this_doc = next(iter(r.docs.values()))
            all_docs.append(this_doc)
            await self.shared_state.docs.aadd_texts(texts=r.texts, doc=this_doc)

        status = await self.shared_state.get_status()

        logger.info(status)

        # mark how far we've searched so that continuation will start at the right place
        self.previous_searches[search_key] += self.search_count

        if self.return_paper_metadata:
            retrieved_papers = "\n".join([f"{x.title} ({x.year})" for x in all_docs])
            return f"Retrieved Papers:\n{retrieved_papers}\n\n{status}"
        return status


class EmptyDocsError(RuntimeError):
    """Error to throw when we needed docs to be present."""


class GatherEvidenceTool(BaseTool):
    class InputSchema(
        BaseModelV1  # TODO: move to Pydantic v2 after LangChain moves to Pydantic v2, SEE: https://github.com/langchain-ai/langchain/issues/16306
    ):
        question: str = FieldV1(description="Specific question to gather evidence for.")

    name: str = "gather_evidence"
    args_schema: type[BaseModelV1] | None = InputSchema
    description: str = (
        "Gather evidence from previous papers given a specific question. "
        "This will increase evidence and relevant paper counts. "
        "Only invoke when paper count is above zero."
    )

    shared_state: SharedToolState
    query: QueryRequest

    def _run(self, query: str) -> str:
        raise NotImplementedError

    @_time_tool
    async def _arun(self, question: str) -> str:
        if not self.shared_state.docs.docs:
            raise EmptyDocsError("Not gathering evidence due to having no papers.")

        logger.info(f"Gathering and ranking evidence for '{question}'.")

        # first we see if we'd like to filter any docs for relevance
        # at the citation level
        if len(self.shared_state.docs.docs) >= self.query.adoc_match_threshold:
            doc_keys_to_keep = await self.shared_state.docs.adoc_match(
                question,
                rerank=True,  # want to set it explicitly
                answer=self.shared_state.answer,
            )
        else:
            doc_keys_to_keep = set(self.shared_state.docs.docs.keys())

        self.shared_state.answer.dockey_filter = doc_keys_to_keep

        # swap out the question
        # TODO: evaluate how often does the agent changes the question
        old = self.shared_state.answer.question
        self.shared_state.answer.question = question

        # generator, so run it
        l0 = len(self.shared_state.answer.contexts)

        # set jit so that the index is rebuilt; helps if the texts have changed
        self.shared_state.docs.jit_texts_index = True
        # ensure length is set correctly
        self.shared_state.answer.summary_length = self.query.summary_length
        # TODO: refactor answer out of this...
        self.shared_state.answer = await self.shared_state.docs.aget_evidence(
            answer=self.shared_state.answer,
            max_sources=self.query.max_sources,
            k=self.query.consider_sources,
            detailed_citations=True,
        )
        l1 = len(self.shared_state.answer.contexts)
        self.shared_state.answer.question = old
        sorted_contexts = sorted(
            self.shared_state.answer.contexts, key=lambda x: x.score, reverse=True
        )
        best_evidence = ""
        if len(sorted_contexts) > 0:
            best_evidence = f" Best evidence:\n\n{sorted_contexts[0].context}"
        status = await self.shared_state.get_status()
        logger.info(status)
        return f"Added {l1 - l0} pieces of evidence.{best_evidence}\n\n" + status


class GenerateAnswerTool(BaseTool):
    class InputSchema(
        BaseModelV1  # TODO: move to Pydantic v2 after LangChain moves to Pydantic v2, SEE: https://github.com/langchain-ai/langchain/issues/16306
    ):
        question: str = FieldV1(description="Question to be answered.")

    name: str = "gen_answer"
    args_schema: type[BaseModelV1] | None = InputSchema
    description: str = (
        "Ask a model to propose an answer answer using current evidence. "
        "The tool may fail, "
        "indicating that better or different evidence should be found. "
        "Having more than one piece of evidence or relevant papers is best."
    )
    shared_state: SharedToolState
    wipe_context_on_answer_failure: bool = True
    query: QueryRequest

    FAILED_TO_ANSWER: ClassVar[str] = "Failed to answer question."

    @classmethod
    def did_not_fail_to_answer(cls, message: str) -> bool:
        return not message.startswith(cls.FAILED_TO_ANSWER)

    def _run(self, query: str) -> str:
        raise NotImplementedError

    @_time_tool
    async def _arun(self, question: str) -> str:
        logger.info(f"Generating answer for '{question}'.")
        # TODO: Should we allow the agent to change the question?
        # self.answer.question = query
        self.shared_state.answer.answer_length = self.query.length
        self.shared_state.answer = await self.shared_state.docs.aquery(
            self.query.query,
            k=self.query.consider_sources,
            max_sources=self.query.max_sources,
            answer=self.shared_state.answer,
        )

        if self.query.filter_extra_background:
            # filter out "(Extra Background Information)" from the answer
            self.shared_state.answer.answer = re.sub(
                r"\([Ee]xtra [Bb]ackground [Ii]nformation\)",
                "",
                self.shared_state.answer.answer,
            )

        if "cannot answer" in self.shared_state.answer.answer.lower():
            if self.wipe_context_on_answer_failure:
                self.shared_state.answer.contexts = []
                self.shared_state.answer.dockey_filter = None
                self.shared_state.answer.context = ""
            status = await self.shared_state.get_status()
            logger.info(status)
            return f"{self.FAILED_TO_ANSWER} | " + status
        status = await self.shared_state.get_status()
        logger.info(status)
        return f"{self.shared_state.answer.answer} | {status}"

    # NOTE: can match failure to answer or an actual answer
    ANSWER_SPLIT_REGEX_PATTERN: ClassVar[str] = (
        r" \| " + SharedToolState.STATUS_SEARCH_REGEX_PATTERN
    )

    @classmethod
    def extract_answer_from_message(cls, content: str) -> str:
        """Extract the answer from a message content."""
        answer, *rest = re.split(
            pattern=cls.ANSWER_SPLIT_REGEX_PATTERN, string=content, maxsplit=1
        )
        if len(rest) != 4 or not cls.did_not_fail_to_answer(answer):  # noqa: PLR2004
            return ""
        return answer


AVAILABLE_TOOL_NAME_TO_CLASS: dict[str, type[BaseTool]] = {
    cls.__fields__["name"].default: cls
    for _, cls in inspect.getmembers(
        sys.modules[__name__],
        predicate=lambda v: inspect.isclass(v)
        and issubclass(v, BaseTool)
        and v is not BaseTool,
    )
}


def query_to_tools(
    query: QueryRequest,
    state: SharedToolState,
    callbacks: list[BaseCallbackHandler] | None = None,
) -> list[BaseTool]:
    if query.agent_tools.tool_names is None:
        tool_types: list[type[BaseTool]] = [
            PaperSearchTool,
            GatherEvidenceTool,
            GenerateAnswerTool,
        ]
    else:
        tool_types = [
            AVAILABLE_TOOL_NAME_TO_CLASS[name]
            for name in set(query.agent_tools.tool_names)
        ]
    tools: list[BaseTool] = []
    for tool_type in tool_types:
        if issubclass(tool_type, PaperSearchTool):
            tools.append(
                PaperSearchTool(
                    shared_state=state,
                    search_count=query.agent_tools.search_count,
                    embedding=query.embedding,
                    parsing_configuration=query.parsing_configuration,
                    paper_directory=query.agent_tools.paper_directory,
                    index_directory=query.agent_tools.index_directory
                    or pqa_directory("indexes"),
                    manifest_file=query.agent_tools.manifest_file,
                    callbacks=callbacks,
                )
            )
        elif issubclass(tool_type, GatherEvidenceTool):
            tools.append(
                GatherEvidenceTool(shared_state=state, query=query, callbacks=callbacks)
            )
        elif issubclass(tool_type, GenerateAnswerTool):
            tools.append(
                GenerateAnswerTool(
                    shared_state=state,
                    wipe_context_on_answer_failure=query.agent_tools.wipe_context_on_answer_failure,
                    query=query,
                    callbacks=callbacks,
                )
            )
        else:
            tools.append(tool_type(shared_state=state))
    return tools



from __future__ import annotations

import asyncio
import logging
import os
import time
from contextlib import asynccontextmanager
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, ClassVar, Collection, assert_never
from uuid import UUID, uuid4

from langchain_core.callbacks import AsyncCallbackHandler
from langchain_core.messages import BaseMessage, messages_to_dict
from langchain_core.outputs import ChatGeneration, LLMResult
from openai import AsyncOpenAI
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)
from typing_extensions import Protocol

from .. import (
    Answer,
    OpenAILLMModel,
    PromptCollection,
    llm_model_factory,
)
from ..utils import hexdigest
from ..version import __version__
from .prompts import STATIC_PROMPTS

logger = logging.getLogger(__name__)


class SupportsPickle(Protocol):
    """Type protocol for typing any object that supports pickling."""

    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __getstate__(self) -> object: ...
    def __setstate__(self, state: object) -> None: ...


class AgentStatus(str, Enum):
    # FAIL - no answer could be generated
    FAIL = "fail"
    # SUCCESS - answer was generated
    SUCCESS = "success"
    # TIMEOUT - agent took too long, but an answer was generated
    TIMEOUT = "timeout"
    # UNSURE - the agent was unsure, but an answer is present
    UNSURE = "unsure"


class AgentPromptCollection(BaseModel):
    """Configuration for the agent."""

    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True)

    agent_system_prompt: str | None = Field(
        # Matching https://github.com/langchain-ai/langchain/blob/langchain%3D%3D0.2.3/libs/langchain/langchain/agents/openai_functions_agent/base.py#L213-L215
        default="You are a helpful AI assistant.",
        description="Optional system prompt message to precede the below agent_prompt.",
    )

    # TODO: make this prompt more minimalist, instead improving tool descriptions so
    # how to use them together can be intuited, and exposing them for configuration
    agent_prompt: str = (
        "Answer question: {question}"
        "\n\nSearch for papers, gather evidence, collect papers cited in evidence then re-gather evidence, and answer."
        " Gathering evidence will do nothing if you have not done a new search or collected new papers."
        " If you do not have enough evidence to generate a good answer, you can:"
        "\n- Search for more papers (preferred)"
        "\n- Collect papers cited by previous evidence (preferred)"
        "\n- Gather more evidence using a different phrase"
        "\nIf you search for more papers or collect new papers cited by previous evidence,"
        " remember to gather evidence again."
        " Once you have five or more pieces of evidence from multiple sources, or you have tried a few times, "
        "call {gen_answer_tool_name} tool. The {gen_answer_tool_name} tool output is visible to the user, "
        "so you do not need to restate the answer and can simply terminate if the answer looks sufficient. "
        "The current status of evidence/papers/cost is {status}"
    )
    paper_directory: str | os.PathLike = Field(
        default=Path.cwd(),
        description=(
            "Local directory which contains the papers to be indexed and searched."
        ),
    )
    index_directory: str | os.PathLike | None = Field(
        default=None,
        description=(
            "Directory to store the PQA generated search index, configuration, and answer indexes."
        ),
    )
    manifest_file: str | os.PathLike | None = Field(
        default=None,
        description=(
            "Optional manifest CSV, containing columns which are attributes for a DocDetails object. "
            "Only 'file_location','doi', and 'title' will be used when indexing."
        ),
    )
    search_count: int = 8
    wipe_context_on_answer_failure: bool = True
    timeout: float = Field(
        default=500.0,
        description=(
            "Matches LangChain AgentExecutor.max_execution_time (seconds), the timeout"
            " on agent execution."
        ),
    )
    should_pre_search: bool = Field(
        default=False,
        description="If set to true, run the search tool before invoking agent.",
    )
    agent_config: dict[str, Any] | None = Field(
        default=None,
        description="Optional keyword argument configuration for the agent.",
    )
    tool_names: Collection[str] | None = Field(
        default=None,
        description=(
            "Optional override on the tools to provide the agent. Leaving as the"
            " default of None will use a minimal toolset of the paper search, gather"
            " evidence, collect cited papers from evidence, and gen answer. If passing tool"
            " names (non-default route), at least the gen answer tool must be supplied."
        ),
    )

    @field_validator("tool_names")
    @classmethod
    def validate_tool_names(cls, v: set[str] | None) -> set[str] | None:
        if v is None:
            return None
        # imported here to avoid circular imports
        from .main import GenerateAnswerTool

        answer_tool_name = GenerateAnswerTool.__fields__["name"].default
        if answer_tool_name not in v:
            raise ValueError(
                f"If using an override, must contain at least the {answer_tool_name}."
            )
        return v


class ParsingOptions(str, Enum):
    PAPERQA_DEFAULT = "paperqa_default"

    def available_for_inference(self) -> list[ParsingOptions]:
        return [self.PAPERQA_DEFAULT]  # type: ignore[list-item]

    def get_parse_type(self, config: ParsingConfiguration) -> str:
        if self == ParsingOptions.PAPERQA_DEFAULT:
            return config.parser_version_string
        assert_never()


class ChunkingOptions(str, Enum):
    SIMPLE_OVERLAP = "simple_overlap"

    @property
    def valid_parsings(self) -> list[ParsingOptions]:
        # Note that SIMPLE_OVERLAP must be valid for all by default
        # TODO: implement for future parsing options
        valid_parsing_dict: dict[str, list[ParsingOptions]] = {}
        return valid_parsing_dict.get(self.value, [])


class ImpossibleParsingError(Exception):
    """Error to throw when a parsing is impossible."""

    LOG_METHOD_NAME: ClassVar[str] = "warning"


class ParsingConfiguration(BaseModel):
    """Holds a superset of params and methods needed for each algorithm."""

    ordered_parser_preferences: list[ParsingOptions] = [
        ParsingOptions.PAPERQA_DEFAULT,
    ]
    chunksize: int = 6000
    overlap: int = 100
    chunking_algorithm: ChunkingOptions = ChunkingOptions.SIMPLE_OVERLAP

    def chunk_type(self, chunking_selection: ChunkingOptions | None = None) -> str:
        """Future chunking implementations (i.e. by section) will get an elif clause here."""
        if chunking_selection is None:
            chunking_selection = self.chunking_algorithm
        if chunking_selection == ChunkingOptions.SIMPLE_OVERLAP:
            return (
                f"{self.parser_version_string}|{chunking_selection.value}"
                f"|tokens={self.chunksize}|overlap={self.overlap}"
            )
        assert_never()

    @property
    def parser_version_string(self) -> str:
        return f"paperqa-{__version__}"

    def is_chunking_valid_for_parsing(self, parsing: str):
        # must map the parsings because they won't include versions by default
        return (
            self.chunking_algorithm == ChunkingOptions.SIMPLE_OVERLAP
            or parsing
            in {  # type: ignore[unreachable]
                p.get_parse_type(self) for p in self.chunking_algorithm.valid_parsings
            }
        )


class MismatchedModelsError(Exception):
    """Error to throw when model clients clash ."""

    LOG_METHOD_NAME: ClassVar[str] = "warning"


class QueryRequest(BaseModel):
    query: str = ""
    id: UUID = Field(
        default_factory=uuid4,
        description="Identifier which will be propagated to the Answer object.",
    )
    llm: str = "gpt-4o-2024-08-06"
    agent_llm: str = Field(
        default="gpt-4o-2024-08-06",
        description="Chat model to use for agent planning",
    )
    summary_llm: str = "gpt-4o-2024-08-06"
    length: str = "about 200 words, but can be longer if necessary"
    summary_length: str = "about 100 words"
    max_sources: int = 10
    consider_sources: int = 16
    named_prompt: str | None = None
    # if you change this to something other than default
    # modify code below in update_prompts
    prompts: PromptCollection = Field(
        default=STATIC_PROMPTS["default"], validate_default=True
    )
    agent_tools: AgentPromptCollection = Field(default_factory=AgentPromptCollection)
    texts_index_mmr_lambda: float = 1.0
    texts_index_embedding_config: dict[str, Any] | None = None
    docs_index_mmr_lambda: float = 1.0
    docs_index_embedding_config: dict[str, Any] | None = None
    parsing_configuration: ParsingConfiguration = ParsingConfiguration()
    embedding: str = "text-embedding-3-small"
    # concurrent number of summary calls to use inside Doc object
    max_concurrent: int = 20
    temperature: float = 0.0
    summary_temperature: float = 0.0
    # at what size should we start using adoc_match?
    adoc_match_threshold: int = 500
    # Should we filter out "Extra Background Information" citations
    # which come from pre-step in paper-qa algorithm
    filter_extra_background: bool = True
    # provides post-hoc linkage of request to a docs object
    # NOTE: this isn't a unique field, on the user to keep straight
    _docs_name: str | None = PrivateAttr(default=None)

    # strict validation for now
    model_config = ConfigDict(extra="forbid")

    @computed_field  # type: ignore[misc]
    @property
    def docs_name(self) -> str | None:
        return self._docs_name

    @model_validator(mode="after")
    def llm_models_must_match(self) -> QueryRequest:
        llm = llm_model_factory(self.llm)
        summary_llm = llm_model_factory(self.summary_llm)
        if type(llm) is not type(summary_llm):
            raise MismatchedModelsError(
                f"Answer LLM and summary LLM types must match: {type(llm)} != {type(summary_llm)}"
            )
        return self

    @field_validator("prompts")
    def update_prompts(
        cls,  # noqa: N805
        v: PromptCollection,
        info: ValidationInfo,
    ) -> PromptCollection:
        values = info.data
        if values["named_prompt"] is not None:
            if values["named_prompt"] not in STATIC_PROMPTS:
                raise ValueError(
                    f"Named prompt {values['named_prompt']} not in {list(STATIC_PROMPTS.keys())}"
                )
            v = STATIC_PROMPTS[values["named_prompt"]]
        if values["summary_llm"] == "none":
            v.skip_summary = True
            # for simplicity (it is not used anywhere)
            # so that Docs doesn't break when we don't have a summary_llm
            values["summary_llm"] = "gpt-4o-mini"
        return v

    def set_docs_name(self, docs_name: str) -> None:
        """Set the internal docs name for tracking."""
        self._docs_name = docs_name

    @staticmethod
    def get_index_name(
        paper_directory: str | os.PathLike,
        embedding: str,
        parsing_configuration: ParsingConfiguration,
    ) -> str:

        # index name should use an absolute path
        # this way two different folders where the
        # user locally uses '.' will make different indexes
        if isinstance(paper_directory, Path):
            paper_directory = str(paper_directory.absolute())

        index_fields = "|".join(
            [
                str(paper_directory),  # cast for typing
                embedding,
                str(parsing_configuration.chunksize),
                str(parsing_configuration.overlap),
                parsing_configuration.chunking_algorithm,
            ]
        )

        return f"pqa_index_{hexdigest(index_fields)}"


class AnswerResponse(BaseModel):
    answer: Answer
    usage: dict[str, list[int]]
    bibtex: dict[str, str] | None = None
    status: AgentStatus
    timing_info: dict[str, dict[str, float]] | None = None
    duration: float = 0.0
    # A placeholder for interesting statistics we can show users
    # about the answer, such as the number of sources used, etc.
    stats: dict[str, str] | None = None

    @field_validator("answer")
    def strip_answer(
        cls, v: Answer, info: ValidationInfo  # noqa: ARG002, N805
    ) -> Answer:
        # This modifies in place, this is fine
        # because when a response is being constructed,
        # we should be done with the Answer object
        v.filter_content_for_user()
        return v

    async def get_summary(self, llm_model="gpt-4-turbo") -> str:
        sys_prompt = (
            "Revise the answer to a question to be a concise SMS message. "
            "Use abbreviations or emojis if necessary."
        )
        model = OpenAILLMModel(config={"model": llm_model, "temperature": 0.1})
        chain = model.make_chain(
            AsyncOpenAI(), prompt="{question}\n\n{answer}", system_prompt=sys_prompt
        )
        result = await chain({"question": self.answer.question, "answer": self.answer.answer})  # type: ignore[call-arg]
        return result.text.strip()


class TimerData(BaseModel):
    start_time: float = Field(default_factory=time.time)  # noqa: FURB111
    durations: list[float] = Field(default_factory=list)


class SimpleProfiler(BaseModel):
    """Basic profiler with start/stop and named timers.

    The format for this logger needs to be strictly followed, as downstream google
    cloud monitoring is based on the following
    # [Profiling] {**name** of timer} | {**elapsed** time of function} | {**__version__** of PaperQA}
    """

    timers: dict[str, list[float]] = {}
    running_timers: dict[str, TimerData] = {}
    uid: UUID = Field(default_factory=uuid4)

    @asynccontextmanager
    async def timer(self, name: str):
        start_time = asyncio.get_running_loop().time()
        try:
            yield
        finally:
            end_time = asyncio.get_running_loop().time()
            elapsed = end_time - start_time
            self.timers.setdefault(name, []).append(elapsed)
            logger.info(
                f"[Profiling] | UUID: {self.uid} | NAME: {name} | TIME: {elapsed:.3f}s | VERSION: {__version__}"
            )

    def start(self, name: str) -> None:
        try:
            self.running_timers[name] = TimerData()
        except RuntimeError:  # No running event loop (not in async)
            self.running_timers[name] = TimerData(start_time=time.time())

    def stop(self, name: str):
        timer_data = self.running_timers.pop(name, None)
        if timer_data:
            try:
                t_stop: float = asyncio.get_running_loop().time()
            except RuntimeError:  # No running event loop (not in async)
                t_stop = time.time()
            elapsed = t_stop - timer_data.start_time
            self.timers.setdefault(name, []).append(elapsed)
            logger.info(
                f"[Profiling] | UUID: {self.uid} | NAME: {name} | TIME: {elapsed:.3f}s | VERSION: {__version__}"
            )
        else:
            logger.warning(f"Timer {name} not running")

    def results(self) -> dict[str, dict[str, float]]:
        result = {}
        for name, durations in self.timers.items():
            mean = sum(durations) / len(durations)
            result[name] = {
                "low": min(durations),
                "mean": mean,
                "max": max(durations),
                "total": sum(durations),
            }
        return result


class AgentCallback(AsyncCallbackHandler):
    """
    Callback handler used to monitor the agent, for debugging.

    Its various capabilities include:
    - Chain start --> error/stop: profile runtime
    - Tool start: count tool invocations
    - LLM start --> error/stop: insert into LLMResultDB

    NOTE: this is not a thread safe implementation since start(s)/end(s) mutate self.
    """

    def __init__(
        self, profiler: SimpleProfiler, name: str, answer_id: UUID, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.profiler = profiler
        self.name = name
        self._tool_starts: list[str] = []
        self._answer_id = answer_id
        # This will be None before/after a completion, and a dict during one
        self._llm_result_db_kwargs: dict[str, Any] | None = None

    @property
    def tool_invocations(self) -> list[str]:
        return self._tool_starts

    async def on_chain_start(self, *args, **kwargs) -> None:
        await super().on_chain_start(*args, **kwargs)
        self.profiler.start(self.name)

    async def on_chain_end(self, *args, **kwargs) -> None:
        await super().on_chain_end(*args, **kwargs)
        self.profiler.stop(self.name)

    async def on_chain_error(self, *args, **kwargs) -> None:
        await super().on_chain_error(*args, **kwargs)
        self.profiler.stop(self.name)

    async def on_tool_start(
        self, serialized: dict[str, Any], input_str: str, **kwargs
    ) -> None:
        await super().on_tool_start(serialized, input_str, **kwargs)
        self._tool_starts.append(serialized["name"])

    async def on_chat_model_start(
        self,
        serialized: dict[str, Any],  # noqa: ARG002
        messages: list[list[BaseMessage]],
        **kwargs,
    ) -> None:
        # NOTE: don't call super(), as it changes semantics
        if len(messages) != 1:
            raise NotImplementedError(f"Didn't handle shape of messages {messages}.")
        self._llm_result_db_kwargs = {
            "answer_id": self._answer_id,
            "name": f"tool_selection:{len(messages[0])}",
            "prompt": {
                "messages": messages_to_dict(messages[0]),
                # SEE: https://platform.openai.com/docs/api-reference/chat/create#chat-create-functions
                "functions": kwargs["invocation_params"]["functions"],
                "tool_history": self.tool_invocations,
            },
            "model": kwargs["invocation_params"]["model"],
            "date": datetime.now().isoformat(),
        }

    async def on_llm_end(self, response: LLMResult, **kwargs) -> None:
        await super().on_llm_end(response, **kwargs)
        if (
            len(response.generations) != 1
            or len(response.generations[0]) != 1
            or not isinstance(response.generations[0][0], ChatGeneration)
        ):
            raise NotImplementedError(
                f"Didn't handle shape of generations {response.generations}."
            )
        if self._llm_result_db_kwargs is None:
            raise NotImplementedError(
                "There should have been an LLM result populated here by now."
            )
        if not isinstance(response.llm_output, dict):
            raise NotImplementedError(
                f"Expected llm_output to be a dict, but got {response.llm_output}."
            )
